@@ -5,10 +5,13 @@ from .utils import CommandUtils, BLEInitializationError
 from .config import Settings
 from .cloud import Cloud
 from .ble import Ble
+import asyncio
+import logging
 
 
 class AiraHome:
     def __init__(self,
+                 ext_loop: asyncio.AbstractEventLoop | None = None,
                  user_pool_id: str = Settings.USER_POOL_IDS[0],
                  client_id: str = Settings.CLIENT_ID,
                  aira_backend: str = Settings.AIRA_BACKEND,
@@ -36,9 +39,16 @@ class AiraHome:
         self.ble_notify_timeout = ble_notify_timeout
         self.max_ble_chunk_size = max_ble_chunk_size
         
+        # Setup logger with NullHandler (no output by default)
+        self.logger = logging.getLogger('pyairahome')
+        if not self.logger.handlers:
+            self.logger.addHandler(logging.NullHandler())
+        self.logger.setLevel(logging.DEBUG)  # Allow all levels, let handlers decide
+        
         # Initialize cloud instance with reference to this class as parent
         self._cloud = None
         # Initialize ble instance with reference to this class as parent
+        self.ext_loop = ext_loop
         self._ble = None
 
         # Store data needed for simple ble usage
@@ -46,7 +56,17 @@ class AiraHome:
         self.uuid = None
 
         # Utils
-        self.command_list = self.get_command_list()
+        self._command_list = None
+        
+        self.logger.info("AiraHome instance initialized")
+
+    @property
+    def command_list(self):
+        """Get the list of available commands."""
+        if not self._command_list:
+            self._command_list = self.get_command_list()
+        
+        return self._command_list
 
     @property
     def cloud(self):
@@ -59,7 +79,7 @@ class AiraHome:
     def ble(self):
         """Get the Ble instance with access to parent AiraHome methods."""
         if self._ble is None:
-            self._ble = Ble(self)
+            self._ble = Ble(self, ext_loop=self.ext_loop)
         return self._ble
 
     ###
@@ -81,19 +101,44 @@ class AiraHome:
     
     def init_ble(self) -> bool:
         """Initialize BLE by fetching the certificate and UUID from the cloud."""
-        if not self.certificate or not self.uuid:
-            devices = self.cloud.get_devices(raw=False)
-            if len(devices["devices"])-1 < self.default_uuid_selection:
-                raise BLEInitializationError(f"Default UUID selection index {self.default_uuid_selection} is out of range for available devices ({len(devices['devices'])}). Please adjust using default_uuid_selection parameter when initiating AiraHome class.")
-             
-            device = devices["devices"][self.default_uuid_selection]
-            self.uuid = device["id"]["value"]
-
-            device_details = self.cloud.get_device_details(self.uuid, raw=False)
-
-            self.certificate = device_details["heat_pump"]["certificate"]["certificate_pem"]
-
-            # try connecting
-            return self.ble.connect()
+        self.logger.info("Initializing BLE connection")
         
-        return False
\ No newline at end of file
+        if not self.certificate or not self.uuid:
+            self.logger.debug("Certificate or UUID not available, fetching from cloud")
+            try:
+                devices = self.cloud.get_devices(raw=False)
+                if len(devices["devices"])-1 < self.default_uuid_selection:
+                    error_msg = f"Default UUID selection index {self.default_uuid_selection} is out of range for available devices ({len(devices['devices'])}). Please adjust using default_uuid_selection parameter when initiating AiraHome class."
+                    self.logger.error(error_msg)
+                    raise BLEInitializationError(error_msg)
+                 
+                device = devices["devices"][self.default_uuid_selection]
+                self.uuid = device["id"]["value"]
+                self.logger.debug(f"Selected device UUID: {self.uuid}")
+
+                device_details = self.cloud.get_device_details(self.uuid, raw=False)
+                self.certificate = device_details["heat_pump"]["certificate"]["certificate_pem"]
+                self.logger.debug("Certificate retrieved from cloud")
+
+                # try connecting
+                connection_result = self.ble.connect()
+                if connection_result:
+                    self.logger.info("BLE connection established successfully")
+                else:
+                    self.logger.warning("BLE connection failed")
+                return connection_result
+            except Exception as e:
+                self.logger.error(f"Failed to initialize BLE: {e}", exc_info=True)
+                raise
+        else:
+            self.logger.debug("Certificate and UUID already available, attempting direct connection")
+            try:
+                connection_result = self.ble.connect()
+                if connection_result:
+                    self.logger.info("BLE connection established successfully")
+                else:
+                    self.logger.warning("BLE connection failed")
+                return connection_result
+            except Exception as e:
+                self.logger.error(f"Failed to connect BLE: {e}", exc_info=True)
+                raise
\ No newline at end of file
